# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing replication group specific operations.

ReplicationGroups and ReplicationGroup are 2 classes defined in this file.

ReplicationGroups:     Class for representing all the replication groups associated
                        with a specific client

ReplicationGroup:      Class for a single replication group selected for a client,
                        and to perform operations on that replication group


ReplicationGroups:
    ReplicationGroupType                        --  Enum to denote all possible types of replication groups
    __init__(commcell_object)                   --  Initialise object of ReplicationGroups class

    __str__()                                   --  Returns all the replication groups

    __repr__()                                  --  Returns the string for the instance of the
                                                    ReplicationGroups class

    has_replication_group(
            replication_group_name)             --  Checks if replication group exists with the given name

    get(replication_group_name)                 --  Returns the ReplicationGroup class object of the input
                                                    replication name

    delete(replication_group_name)              --  Delete replication group with replication group name

    refresh()                                   --  Refresh all replication groups created on the commcell

    #### internal methods ###
    _get_replication_groups()                   --  REST API call to get all replication groups
                                                    in the commcell

    ##### properties ######
    replication_groups()                        --  Returns all replication groups in the commcell


ReplicationGroup:
    __init__(commcell_object,
            replication_group_name)                 -- Initialise object of ReplicationGroup with the
                                                        specified replication group name

    __repr__()                                      -- return the ReplicationGroup name

    refresh()                                       -- Refresh the object properties

    ##### internal methods #####
    _get_replication_group_dict()                   -- Method to get replication group dictionary

    _get_replication_group_properties()             -- Get the replication group properties

    ##### properties #####
    group_name()                                    -- Returns the replication group name
    group_id()                                      -- Returns the replication group ID
    task_id()                                       -- Returns the ID of the task associated to replication group
    replication_type()                              -- Returns the enum constant of the ReplicationGroupType
    zeal_group()                                    -- Returns a boolean to denote whether group is Zeal
                                                        or backup-based

    restore_options()                               -- Returns a hypervisor specific set of restore options

    is_dvdf_enabled()                               -- Returns whether 'Deploy VM during failover' enabled
    is_warm_sync_enabled()                          -- Returns whether 'Warm site recovery' is enabled

    source_client()                                 -- Returns a client object of the source hypervisor
    destination_client()                            -- Returns a client object of the destination hypervisor

    source_agent()                                  -- Returns an agent object of the source client
    destination_agent()                             -- Returns an agent object of the destination client

    source_instance()                               -- Returns an instance object of the source agent
    destination_instance()                          -- Returns an instance object of the destination agent

    subclient()                                     -- Returns the subclient object of the VM group associated
                                                        with the replication group

    live_sync_pairs()                               -- Returns the list of source VM names that are already present in
                                                        replication monitor
    vm_pairs()                                      -- Returns a dictionary of source VM names
                                                        and LiveSyncVMPairs object mapping

    is_enabled()                                    -- Returns a boolean to tell whether replication group
                                                        is enabled or disabled
    group_frequency()                               -- Returns the group frequency in minutes

    copy_precedence_applicable                      -- Returns a boolean whether the copy precedence is applicable or
                                                        not

    copy_for_replication()                          -- Returns the copy precedence ID used for replication

"""

from __future__ import absolute_import
from __future__ import unicode_literals

from enum import Enum

from cvpysdk.drorchestration.blr_pairs import BLRPairs
from past.builtins import basestring
from ..exception import SDKException


class ReplicationGroups:
    """Class for getting all the replication groups in commcell."""

    class ReplicationGroupType(Enum):
        """ Enum to map Replication Group Types to integers"""
        VSA_PERIODIC = 1
        VSA_CONTINUOUS = 2
        FILE_SYSTEM = 3
        ORACLE = 4
        SQL_SERVER = 5
        SAP_HANA = 6

    def __init__(self, commcell_object):
        """Initialize object of the Replication groups
            Args:
                commcell_object (Commcell)  --  instance of the Commcell class
        """
        self._commcell_object = commcell_object
        self._services = commcell_object._services

        self._replication_groups = None

        self.refresh()

    def __str__(self):
        """Representation string consisting of all replication groups in a formatted output
            Returns:
                str - string of all the replication groups
        """
        representation_string = '{:^5}\t{:^20}\t{:^20}\n\n'.format(
            'S. No.', 'Replication Group Id', 'Replication Group')

        for index, replication_group in enumerate(self._replication_groups):
            sub_str = '{:^5}\t{:20}\t{:20}\n'.format(
                index + 1,
                self._replication_groups[replication_group],
                replication_group
            )
            representation_string += sub_str

        return representation_string.strip()

    def __repr__(self):
        """Representation string for the instance of the ReplicationGroups class."""
        return "Replication Groups for Commserv: '{0}'".format(
            self._commcell_object.commserv_name)

    def has_replication_group(self, replication_group_name):
        """Checks if replication group exists or not

            Args:
                replication_group_name (str)  --  name of the replication group

            Returns:
                bool - boolean output whether replication group exists or not

            Raises:
                SDKException:
                    if proper inputs are not provided
        """
        if not isinstance(replication_group_name, basestring):
            raise SDKException('ReplicationGroup', '101')

        return self.replication_groups and replication_group_name.lower() in self.replication_groups

    def get(self, replication_group_name):
        """Returns a replication group object of the specified replication group name.

            Args:
                replication_group_name (str)  --  name of the replication group

            Returns:
                object - instance of the ReplicationGroup class for the given replication group name

            Raises:
                SDKException:
                    if proper inputs are not provided
                    If Replication group doesnt exists with given name
        """
        if not isinstance(replication_group_name, basestring):
            raise SDKException('ReplicationGroup', '101')
        replication_group_name = replication_group_name.lower()
        if self.has_replication_group(replication_group_name):
            return ReplicationGroup(
                self._commcell_object, replication_group_name)

        raise SDKException(
            'ReplicationGroup',
            '102',
            "Replication group doesn't exist with name: {0}".format(replication_group_name))

    def delete(self, replication_group_name):
        """ Deletes the specified replication group name.

            Args:
                replication_group_name (str)  --  name of the replication group

            Returns:


            Raises:
                SDKException:
                    if proper inputs are not provided
                    if response is empty
                    if response is not success
        """

        if not isinstance(replication_group_name, basestring):
            raise SDKException('ReplicationGroup', '101')

        replication_group_name = replication_group_name.lower()
        if self.has_replication_group(replication_group_name):

            replication_group_dict = self.replication_groups.get(
                replication_group_name.lower(), {})

            if replication_group_dict:
                if replication_group_dict.get('zealGroup'):
                    payload = {
                        "repGrpIds": [int(replication_group_dict.get('id'))],
                        "taskIds": [],
                    }
                else:
                    payload = {
                        "repGrpIds": [],
                        "taskIds": [replication_group_dict.get('id')],
                    }

                flag, response = self._commcell_object._cvpysdk_object.make_request(
                    method='POST',
                    url=self._services['DELETE_REPLICATION_GROUPS'],
                    payload=payload
                )

                if flag:
                    if response.json() and 'deleteGroupsResponse' in response.json():
                        if (response.json().get('deleteGroupsResponse', [{}])[0]
                                .get('response').get('errorMessage')):
                            error_message = (response.json().get('deleteGroupsResponse', {})
                                             .get('response', {}).get('errorMessage'))
                            o_str = ('Failed to delete replication group: {0} \nError: "{1}"'
                                     .format(replication_group_name, error_message))

                            raise SDKException('ReplicationGroup', '102', o_str)
                        self.refresh()

                    else:
                        raise SDKException('Response', '102')
                else:
                    response_string = self._commcell_object._update_response_(
                        response.text)
                    raise SDKException('Response', '101', response_string)
            else:
                raise SDKException('ReplicationGroup', '101', 'Replication group information is empty')
        else:
            raise SDKException(
                'ReplicationGroup', '102', 'No replication group exists with name: "{0}"'.format(
                    replication_group_name)
            )

    @property
    def replication_groups(self):
        """ return all replication groups
        Args:

        Returns: All the replication groups in the commcell

        Raises:
        """
        return self._replication_groups

    def _get_replication_groups(self):
        """REST API call for all the replication groups in the commcell.
            Args:

            Returns:
                dict - consists of all replication groups
                    {
                         "replication_group_name1": {id: '1', 'type': VSA_PERIODIC, 'zealGroup': true},
                         "replication_group_name2": {id: '2', 'type': VSA_CONTINUOUS, 'zealGroup': false}
                    }

            Raises:
                SDKException:
                    if response is empty
                    if response is not success
        """
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._services['REPLICATION_GROUPS'])

        if flag:
            if response.json() and 'replicationGroups' in response.json():

                replication_groups = {}

                for group_dict in response.json()['replicationGroups']:
                    group_type = (self.ReplicationGroupType(group_dict.get('type'))
                                  if group_dict.get('type') else None)
                    if group_dict.get('replicationGroup', {}).get('replicationGroupId'):
                        group_name = group_dict.get('replicationGroup', {}).get('replicationGroupName', '')
                        group_id = str(group_dict.get('replicationGroup', {}).get('replicationGroupId', 0))
                        zeal_group = True
                    else:
                        subtask_dict = group_dict.get('taskDetail', {}).get('subTasks')[0]
                        group_name = subtask_dict.get('subTask', {}).get('subTaskName', '')
                        group_id = str(group_dict.get('taskDetail', {}).get('task', {}).get('taskId', 0))
                        zeal_group = False
                    replication_groups[group_name.lower()] = {
                        'id': group_id,
                        'type': group_type,
                        'zealGroup': zeal_group,
                    }
                return replication_groups
            raise SDKException('Response', 102)

        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def refresh(self):
        """ Refresh the replication groups created in the commcell.
        Args:

        Returns:

        Raises:

        """
        self._replication_groups = self._get_replication_groups()


class ReplicationGroup:
    """Class for all Replication groups related SDK"""

    def __init__(self, commcell_object, replication_group_name):
        """Initialise the ReplicationGroup object for the given group name
            Args:
                commcell_object (Commcell)      --  instance of the Commcell class
                replication_group_name (str)    --  name of the replication group
        """
        self._commcell_object = commcell_object
        self._services = commcell_object._services
        self._replication_group_properties = None

        self._group_name = replication_group_name.lower()
        self._group_dict = self._get_replication_group_dict()

        self._source_client = None
        self._destination_client = None
        self._source_agent = None
        self._destination_agent = None
        self._source_instance = None
        self._destination_instance = None

        self._subclient = None
        self._vm_pairs = None

        self.refresh()

    def __repr__(self):
        """String representation of the instance of the replication group"""
        representation_string = f'ReplicationGroup class instance for ' \
                                f'{"Zeal" if self.zeal_group else "Backup based"}' \
                                f' replication group: "{self.group_name}"'
        return representation_string.format(self.group_name)

    def __str__(self):
        """Strings showing all VM pairs of the replication group in a formatted output
            Returns:
                str - string of all VM pairs
        """
        representation_string = '{:^5}\t{:^20}\t{:^20}\n\n'.format('Pair Id', 'Source VM', 'Destination VM')

        for source_vm in self.vm_pairs:
            sub_str = '{:^5}\t{:20}\t{:20}\n'.format(
                self.vm_pairs[source_vm].vm_pair_id,
                source_vm,
                self.vm_pairs[source_vm].destination_vm
            )
            representation_string += sub_str

        return representation_string.strip()

    def _get_replication_group_dict(self):
        """ Gets replication group dict from the ReplicationGroups class
            Returns: (list) The list of replication groups dictionary objects
        """
        rgs_obj = ReplicationGroups(self._commcell_object)
        return rgs_obj.replication_groups.get(self._group_name)

    def _get_replication_group_properties(self):
        """ Gets replication group properties
            Args:

            Returns: Gets the replication group properties dict

            Raises:
                SDKException:
                    if response is empty

                    if response is not success
        """
        if self.zeal_group:
            flag, response = self._commcell_object._cvpysdk_object.make_request(
                'GET', self._services['REPLICATION_GROUP_DETAILS'] % self.group_id)
        else:
            flag, response = self._commcell_object._cvpysdk_object.make_request(
                'GET', self._services['LIVE_SYNC_DETAILS'] % self.group_id)

        if flag:
            if response.json().get('replicationInfo', {}).get('replicationTargets', {}).get('taskInfo'):
                return response.json().get('replicationInfo', {}).get('replicationTargets', {}).get('taskInfo')[0]
            if response.json().get('taskInfo'):
                return response.json().get('taskInfo')
            if self.replication_type == ReplicationGroups.ReplicationGroupType.VSA_CONTINUOUS:
                return response.json().get('replicationGroupDetails', {}).get('taskDetail')
            raise SDKException('Response', '102')

        response_string = self._commcell_object._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def refresh(self):
        """ Refresh the replication group properties """
        self._replication_group_properties = self._get_replication_group_properties()

    @property
    def group_name(self):
        """Returns: (str) Returns the name of the replication group"""
        return self._group_name

    @property
    def group_id(self):
        """Returns: (str) Returns the ID of the replication group (Zeal)/subtask(classic)"""
        return str(self._group_dict.get('id'))

    @property
    def task_id(self):
        """Returns: (str) Returns the ID of the task associated to the replication group"""
        return str(self._replication_group_properties.get('task', {}).get('taskId'))

    @property
    def replication_type(self):
        """
        Returns: (enum) Returns the type of the replication group.
        """
        return self._group_dict.get('type')

    @property
    def zeal_group(self):
        """Returns: (bool) True, if zeal replication group, false otherwise"""
        return self._group_dict.get('zealGroup', False)

    @property
    def restore_options(self):
        """
        Returns: (dict) The dictionary of restore options of the replication group
            The dictionary structure depends on the vendor
        """
        return (self._replication_group_properties.get('subTasks', [{}])[0]
                .get('options', {}).get('restoreOptions', {}))

    @property
    def is_dvdf_enabled(self):
        """Returns: (bool) Whether deploy VM during failover is enabled or not"""
        return (self.restore_options.get('virtualServerRstOption', {})
                .get('diskLevelVMRestoreOption', {}).get('deployVmWhenFailover', False))

    @property
    def is_warm_sync_enabled(self):
        """Returns: (bool) Whether Warm sync is enabled or not"""
        return (self.restore_options.get('virtualServerRstOption', {})
                .get('diskLevelVMRestoreOption', {}).get('createVmsDuringFailover', False))

    @property
    def source_client(self):
        """Returns:  the client object of the source hypervisor"""
        if not self._source_client:
            client_name = self._replication_group_properties.get('associations', [{}])[0].get('clientName')
            self._source_client = self._commcell_object.clients.get(client_name)
        return self._source_client

    @property
    def destination_client(self):
        """Returns: (str) the client object for the destination hypervisor"""
        if not self._destination_client:
            client_name = (self.restore_options.get('virtualServerRstOption', {})
                           .get('vCenterInstance', {}).get('clientName'))
            self._destination_client = self._commcell_object.clients.get(client_name)
        return self._destination_client

    @property
    def source_agent(self):
        """Returns: the agent object of the source hypervisor"""
        if not self._source_agent:
            agent_name = self._replication_group_properties.get('associations', [{}])[0].get('appName')
            self._source_agent = self.source_client.agents.get(agent_name)
        return self._source_agent

    @property
    def destination_agent(self):
        """Returns: the agent object of the destination hypervisor"""
        if not self._destination_agent:
            agent_name = self._replication_group_properties.get('associations', [{}])[0].get('appName')
            self._destination_agent = self.destination_client.agents.get(agent_name)
        return self._destination_agent

    @property
    def source_instance(self):
        """Returns: (str) The source hypervisor's instance name"""
        if not self._source_instance:
            instance_name = self._replication_group_properties.get('associations', [{}])[0].get('instanceName')
            self._source_instance = self.source_agent.instances.get(instance_name)
        return self._source_instance

    @property
    def destination_instance(self):
        """Returns: (str) The destination hypervisor's instance name"""
        if not self._destination_instance:
            instance_name = (self.restore_options.get('virtualServerRstOption', {})
                             .get('vCenterInstance', {}).get('instanceName'))
            
            # TODO : Depends on DR Layer changes : Workaround used
            instance_name = 'Amazon Web Services' if instance_name == 'Amazon' else instance_name
            
            self._destination_instance = self.destination_agent.instances.get(instance_name)
        return self._destination_instance

    @property
    def subclient(self):
        """Returns: the subclient object of the replication group"""
        if not self._subclient:
            backupset_name = self._replication_group_properties.get('associations', [{}])[0].get('backupsetName')
            backupset = self.source_instance.backupsets.get(backupset_name)
            subclient_name = self._replication_group_properties.get('associations', [{}])[0].get('subclientName')
            self._subclient = backupset.subclients.get(subclient_name)
        return self._subclient

    @property
    def live_sync_pairs(self):
        """
        Returns: A list of all source VM names for which live sync pair exists for a periodic replication group
            eg: ["vm1", "vm2"]
        """
        _live_sync_pairs = []
        if self.replication_type == ReplicationGroups.ReplicationGroupType.VSA_PERIODIC:
            live_sync_name = self.group_name.replace('_ReplicationPlan__ReplicationGroup', '')
            live_sync = self.subclient.live_sync.get(live_sync_name)
            _live_sync_pairs = list(live_sync.vm_pairs)
        elif self.replication_type == ReplicationGroups.ReplicationGroupType.VSA_CONTINUOUS:
            blr_pairs = BLRPairs(self._commcell_object, self.group_name)
            _live_sync_pairs = list(blr_pairs.blr_pairs)
        else:
            raise SDKException('ReplicationGroup', '101', 'Implemented only for replication groups'
                                                          ' of virtual server periodic')
        return _live_sync_pairs

    @property
    def vm_pairs(self):
        """Returns: A dictionary of livesyncVM pairs/BLR pairs object
            eg: {"src_vm1": LiveSyncVMPair, "src_vm2": LiveSyncVMPair}
        """
        if not self._vm_pairs:
            if self.replication_type == ReplicationGroups.ReplicationGroupType.VSA_PERIODIC:
                live_sync_name = self.group_name.replace('_ReplicationPlan__ReplicationGroup', '')
                live_sync = self.subclient.live_sync.get(live_sync_name)
                self._vm_pairs = {source_vm: live_sync.get(source_vm)
                                  for source_vm in live_sync.vm_pairs}
            elif self.replication_type == ReplicationGroups.ReplicationGroupType.VSA_CONTINUOUS:
                blr_pairs = BLRPairs(self._commcell_object, self.group_name)
                self._vm_pairs = {pair_dict.get('sourceName'):
                                  blr_pairs.get(pair_dict.get('sourceName'), pair_dict.get('destinationName'))
                                  for pair_dict in blr_pairs.blr_pairs.values()}
            else:
                raise SDKException('ReplicationGroup', '101', 'Implemented only for replication groups'
                                                              ' of virtual server periodic')
        return self._vm_pairs

    @property
    def is_enabled(self):
        """Returns: (bool) Returns True if state of the replication group 'Enabled' else False"""
        return not self._replication_group_properties.get('task', {}).get('taskFlags', {}).get('disabled', False)

    @property
    def group_frequency(self):
        """Returns: (int) The frequency in minutes at which the group is synced (only applicable for Zeal groups)"""
        return self._replication_group_properties.get('pattern', {}).get('freq_interval', 0)

    @property
    def copy_precedence_applicable(self):
        """Returns: (bool) Whether the copy precedence is applicable or not"""
        return (self.restore_options.get('browseOption', {}).get('mediaOption', {})
                .get('copyPrecedence', {}).get('copyPrecedenceApplicable', False))

    @property
    def copy_for_replication(self):
        """Returns: (int) The ID of the copy used for the replication"""
        # Copy for replication is only applicable is group has copy precedence enabled
        return (self.restore_options.get('browseOption', {}).get('mediaOption', {})
                .get('copyPrecedence', {}).get('copyPrecedence'))

    @property
    def recovery_target(self):
        """Returns: (str) The recovery target used for the replication"""
        return (self.restore_options.get('virtualServerRstOption', {}).get('allocationPolicy', {})
                .get('vmAllocPolicyName'))



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for configuring and monitoring live sync on the amazon subclient.

AmazonLiveSync is the only class defined in this file.

AmazonLiveSync: Class for configuring and monitoring Amazon subclient live sync

AmazonLiveSync:

    configure_live_sync()     -- To configure live sync from supplied parameters

"""
from past.builtins import basestring
from cvpysdk.subclients.virtualserver.livesync.vsa_live_sync import VsaLiveSync
from cvpysdk.exception import SDKException


class AmazonLiveSync(VsaLiveSync):
    """Class for configuring and monitoring Amazon live sync operations"""
    def configure_live_sync(self,
                            schedule_name=None,
                            destination_client=None,
                            proxy_client=None,
                            copy_precedence=0,
                            power_on=True,
                            unconditional_overwrite=False,
                            pattern_dict=None,
                            vm_to_restore=None,
                            restored_vm_name=None,
                            restore_option=None,
                            networkdisplayname=None,
                            region=None,
                            data_center=None,
                            security_groups=None,
                            network=None,
                            volume_type="",
                            ):
        """To configure live

                Args:

                    schedule_name               (str)   -- Name of the Live sync schedule to be created

                    destination_client          (str)   -- Client Name where VM needs to be restored

                    proxy_client                (str)   -- Name of the proxy client to be used

                    copy_precedence             (int)   -- Copy id from which restore needs to be performed
                                                            default: 0

                    power_on                    (bool)  -- To validate destination VM power on and off
                                                            default: True

                    unconditional_overwrite     (bool)  -- To overwrite VM and disk in destination
                                                            default: False

                    vm_to_restore               (list)  -- VM's to be restored

                    restored_vm_name            (str)   -- Name used for the VM when restored

                    restore_option              (dict)  -- Restore options dictionary with advanced options

                    networkdisplayname          (str)   -- Network Display name
                                                    which has used for destination VM

                    region                      (str)   -- region of the destination vm

                    datacenter                  (str)   -- esx host

                    securityGroups              (dict)  -- security groups to associate with replicated vm

                    network                     (dict)  -- network, subnet to associate with replicated vm

                    pattern_dict                (dict)  -- Dictionary to generate the live sync schedule

                        Sample:

                            for after_job_completes :
                            {
                                "freq_type": 'after_job_completes',
                                "active_start_date": date_in_%m/%d/%y (str),
                                "active_start_time": time_in_%H/%S (str),
                                "repeat_days": days_to_repeat (int)
                            }

                            for daily:
                            {
                                 "freq_type": 'daily',
                                 "active_start_time": time_in_%H/%S (str),
                                 "repeat_days": days_to_repeat (int)
                            }

                            for weekly:
                            {
                                 "freq_type": 'weekly',
                                 "active_start_time": time_in_%H/%S (str),
                                 "repeat_weeks": weeks_to_repeat (int)
                                 "weekdays": list of weekdays ['Monday','Tuesday']
                            }

                            for monthly:
                            {
                                 "freq_type": 'monthly',
                                 "active_start_time": time_in_%H/%S (str),
                                 "repeat_months": weeks_to_repeat (int)
                                 "on_day": Day to run schedule (int)
                            }

                            for yearly:
                            {
                                 "active_start_time": time_in_%H/%S (str),
                                 "on_month": month to run schedule (str) January, February...
                                 "on_day": Day to run schedule (int)
                            }

                Returns:
                    object - instance of the Schedule class for this Live sync
                    :param restored_vm_name:
                    :param vm_to_restore:
                    :param pattern_dict:
                    :param unconditional_overwrite:
                    :param power_on:
                    :param copy_precedence:
                    :param proxy_client:
                    :param networkdisplayname:
                    :param destination_client:
                    :param schedule_name:
                    :param security_groups:

                """
        # restore options
        if restore_option is None:
            restore_option = {}

        if vm_to_restore and not isinstance(vm_to_restore, basestring):
            raise SDKException('Subclient', '101')

        if not restored_vm_name:
            restored_vm_name = "LiveSync_"
        restore_option['restore_new_name'] = restored_vm_name

        if copy_precedence:
            restore_option["copy_precedence_applicable"] = True

        if vm_to_restore:
            vm_to_restore = [vm_to_restore]

        # check mandatory input parameters are correct
        if bool(restore_option):
            if not (isinstance(unconditional_overwrite, bool) and
                    isinstance(power_on, bool)):
                raise SDKException('Subclient', '101')

        # set attr for all the option in restore xml from user inputs
        self._subclient_object._set_restore_inputs(
            restore_option,
            vm_to_restore=self._subclient_object._set_vm_to_restore(vm_to_restore),
            vcenter_client=destination_client,
            unconditional_overwrite=unconditional_overwrite,
            power_on=power_on,
            copy_precedence=copy_precedence,
            volume_level_restore=1,
            client_name=proxy_client,
            esx_host=region,
            in_place=False,
            networkDisplayName=networkdisplayname,
            datacenter=data_center,
            securityGroups=security_groups,
            volumeType=volume_type,
            live_sync=True,
            nics=network
        )

        return self._configure_live_sync(schedule_name, restore_option, pattern_dict)


# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Helper file to maintain all the constants for index server and datacube

IndexServerConstants            -       Maintains constants for index server

ContentAnalyzerConstants        -       Maintains constants for Content Analyzer

"""


class IndexServerConstants:
    """Class to maintain all the index server related constants"""

    INDEX_SERVER_IDA_NAME = "Big Data Apps"
    INDEX_SERVER_INSTANCE_NAME = "dynamicIndexInstance"

    ROLE_DATA_ANALYTICS = "Data Analytics"
    ROLE_FILE_SYSTEM_INDEX = "FileSystem Index"
    ROLE_EXCHANGE_INDEX = "Exchange Index"
    ROLE_ONEDRIVE_INDEX = "OneDrive Index"
    ROLE_SYSTEM_DEFAULT = "System Default"

    ENGINE_NAME = "engineName"
    CLOUD_ID = "cloudID"
    ROLES = "version"
    HOST_NAME = "hostName"
    CLOUD_NAME = 'internalCloudName'
    CLIENT_NAME = 'clientName'
    CI_SERVER_URL = 'cIServerURL'
    TYPE = 'type'
    BASE_PORT = 'basePort'
    CLIENT_ID = 'clientId'
    SERVER_TYPE = 'serverType'
    INDEX_SERVER_CLIENT_ID = 'indexServerClientId'

    FSINDEX = 'fsindex_'
    MULTINODE = '_multinode'

    OPERATION_ADD = 1
    OPERATION_DELETE = 2
    OPERATION_EDIT = 3

    DEFAULT_JVM_MAX_MEMORY = '8191'
    DEFAULT_SOLR_PORT = '20000'

    SOLR_PORT_META_INFO = {
        "name": "PORTNO",
        "value": DEFAULT_SOLR_PORT
    }
    SOLR_JVM_META_INFO = {
        "name": "JVMMAXMEMORY",
        "value": DEFAULT_JVM_MAX_MEMORY
    }

    UPDATE_ADD_ROLE = {
        "roleName": "",
        "operationType": OPERATION_ADD}

    REQUEST_JSON = {
        "opType": OPERATION_ADD,
        "type": 1,
        "configureNodes": True,
        "solrCloudInfo": {
            "roles": []
        },
        "cloudNodes": [],
        "cloudInfoEntity": {},
        "cloudMetaInfos": []

    }


class ContentAnalyzerConstants:
    """Class to maintain all the content analyzer cloud related constants"""

    OPERATION_ADD = 1
    OPERATION_DELETE = 2
    OPERATION_EDIT = 3

    REQUEST_JSON = {
        "opType": OPERATION_ADD,
        "type": 2,
        "configureNodes": True,
        "cloudNodes": [],
        "cloudInfoEntity": {},
    }



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing handler operations on a datasource.

Handlers and Handler are 2 classes defined in this file.

Handlers: Class for representing all the Handlers associated with the datasource

Handler: Class for a single Handler of the datasource

Handlers:

    __init__()                  --  initialize object of Handlers class associated with commcell

    __str__()                   --  returns all the handlers associated with the commcell

    __repr__()                  --  returns the string representing instance of the Handlers class

    _get_handlers()             --  gets all the handlers associated with the commcell

    has_handler()               --  checks if a handler exists with the given name or not

    get_properties()            --  gets the properties of the given handler

    add()                       --  adds a new handler to the datasource

    refresh()                   --  refresh the handlers associated with the datasource

    get()                       -- gets the object for the given handler name

    delete()                    -- deletes the given handler name

Handler:

    __init__()                  -- Initialize object for Handler

    get_handler_data()          -- Execute the handler

    share()                     -- Share the handler with user or usergroup

"""


from __future__ import absolute_import
from __future__ import unicode_literals

from past.builtins import basestring

from ..exception import SDKException


class Handlers(object):
    """Class for representing all the handlers associated with the datasource."""

    def __init__(self, datasource_object):
        """Initialize object of the Handlers class.

            Args:
                _datasource_object (object)  --  instance of the datastore class

            Returns:
                object - instance of the Handlers class

        """

        self._datasource_object = datasource_object
        self.commcell_obj = self._datasource_object._commcell_object
        self._create_handler = self.commcell_obj._services['CREATE_HANDLER']
        self._get_handler = self.commcell_obj._services['GET_HANDLERS'] % (
            self._datasource_object.datasource_id
        )

        self._handlers = None
        self.refresh()

    def __str__(self):
        """Representation string consisting of all handlers of the datasource.

            Returns:
                str - string of all the handlers associated with the datasource

        """
        representation_string = '{:^5}\t{:^20}\n\n'.format('S. No.', 'Handler')
        for index, handler in enumerate(self._handlers):
            sub_str = '{:^5}\t{:20}\n'.format(index + 1, handler)
            representation_string += sub_str

        return representation_string.strip()

    def __repr__(self):
        """Representation string for the instance of the Handlers class."""
        return "Handlers class instance for Datasource: '{0}'".format(
            self._datasource_object.datasource_name
        )

    def _get_handlers(self):
        """Gets all the handlers associated with the datasource

            Returns:
                dict - consists of all handlers in the datasource
                    {
                         "handler1_name": dict of handler1 properties,
                         "handler2_name": dict of handler2 properties
                    }

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """

        flag, response = self._datasource_object._commcell_object._cvpysdk_object.make_request(
            'GET', self._get_handler)
        if flag:
            if response.json() and 'handlerInfos' in response.json():
                handlers_dict = {}
                for dictionary in response.json()['handlerInfos']:
                    temp_name = dictionary['handlerName']
                    handlers_dict[temp_name] = dictionary
                return handlers_dict
            raise SDKException('Response', '102')
        response_string = self._datasource_object._commcell_object._update_response_(
            response.text)
        raise SDKException('Response', '101', response_string)

    def has_handler(self, handler_name):
        """Checks if a handler exists in the datasource with the input handler name.

            Args:
                handler_name (str)  --  name of the handler

            Returns:
                bool - boolean output whether the handler exists in the commcell or not

            Raises:
                SDKException:
                    if type of the handler name argument is not string

        """
        if not isinstance(handler_name, basestring):
            raise SDKException('Datacube', '101')

        return self._handlers and handler_name.lower() in map(str.lower, self._handlers)

    def get_properties(self, handler_name):
        """Returns a handler object of the specified handler name.

            Args:
                handler_name (str)  --  name of the handler

            Returns:
                dict -  properties for the given handler name


        """
        return self._handlers[handler_name]

    def get(self, handler_name):
        """Returns a handler object of the specified handler name.

            Args:
                handler_name (str)  --  name of the handler

            Returns:

                obj                 -- Object of Handler class


        """
        if not isinstance(handler_name, basestring):
            raise SDKException('Datacube', '101')

        if self.has_handler(handler_name):
            handler_id = self.get_properties(handler_name)['handlerId']
            return Handler(self._datasource_object, handler_name, handler_id)
        raise SDKException('Datacube', '102', "Unable to get handler class object")

    def delete(self, handler_name):
        """ deletes the handler associated with this handler object
                Args:

                    handler_name (str)     -- Name of the handler which needs to be deleted

                Returns:

                    None

                Raises:

                    SDKExpception:

                        if response is empty

                        if response is not success
        """
        handler_id = self.get(handler_name).handler_id
        self._delete_handler = self.commcell_obj._services['DELETE_HANDLER'] % (
            handler_id)

        flag, response = self.commcell_obj._cvpysdk_object.make_request(
            'POST', self._delete_handler)

        if flag:
            if 'errorCode' in response.json() and response.json()['errorCode'] != 0:
                error_message = response.json()['errLogMessage']
                o_str = 'Failed to Delete handler on datasource\nError: "{0}"'.format(error_message)
                raise SDKException('Datacube', '102', o_str)
            elif 'errorCode' in response.json() and response.json()['errorCode'] == 0:
                return
            else:
                raise SDKException('Datacube', '102', "Empty Response with no errorCode")
        raise SDKException('Response', '101', response.text)

    def add(self,
            handler_name,
            search_query,
            filter_query=None,
            facet_field=None,
            facet_query=None,
            rows=10,
            response_type="json",
            sort_column=[]):
        """Adds a new handler to the commcell.

            Args:
                handler_name    (str)   --  name of the handler to add to the datastore

                search_query    (list)  --  list of keywords on which the search is performed.

                filter_query    (list)  --  list of conditional queries which needs to be performed
                                                when the data is retrieved

                facet_field     (list)  --  list of fields to be faceted.

                facet_query     (list)  --  list of conditional queries for which the facet count
                                                should be retrieved

                rows            (int)   --  list of keywords on which the search is performed.

                response_type   (str)   --  format in which search results are retrieved.
                    default: json

                    Supported Types:
                        json

                        csv

                        xml


                sort_column     (list)  --  list of column name to be sorted


            Raises:
                SDKException:
                    if type of the handler name argument is not string

                    if failed to delete handler

                    if response is empty

                    if response is not success

                    if no handler exists with the given name

        """
        request_json = {
            "dataSourceId": self._datasource_object.datasource_id,
            "handlerName": handler_name,
            "handlerInfo": {
                "defaultParams": {
                    "q": search_query,
                    "fq": filter_query,
                    "sort": sort_column,
                    "facet": [
                        "true" if facet_field or facet_query else "false"
                    ],
                    "facet.field": facet_field,
                    "facet.query": facet_query,
                    "rows": [rows],
                    "wt": [response_type]
                },
                "appendParams": {},
                "rawDefaultParams": [],
                "rawAppendParams": [],
                "rawInvariantParams": [],
                "alwaysDecode": "true"
            }
        }

        flag, response = self.commcell_obj._cvpysdk_object.make_request(
            'POST', self._create_handler, request_json
        )
        if flag:
            if response.json() and 'error' in response.json():
                if(response.json()['error'] == 'None' or response.json()['error'] is None):
                    self.refresh()  # reload new list.
                    return
                error_code = response.json()['error']['errorCode']
                if error_code == 0:
                    self.refresh()  # reload new list.
                    return
                error_message = response.json()['error']['errLogMessage']
                o_str = 'Failed to create handler\nError: "{0}"'.format(error_message)
                raise SDKException('Response', '102', o_str)
            raise SDKException('Response', '102')
        response_string = self.commcell_obj._update_response_(
            response.text)
        raise SDKException('Response', '101', response_string)

    def refresh(self):
        """Refresh the handlers associated with the Datasource."""
        self._handlers = self._get_handlers()


class Handler(object):
    """Class for performing operations on a single Handler"""

    def __init__(self, datasource_object, handler_name, handler_id=None):
        """Initialize an object of the Handler class.

            Args:
                datasource_object     (object)    --  instance of the Datacube class

                handler_name          (str)       --  name of the Handler

                handler_id            (int)       --  Id of the Handler. Default = None

            Returns:

                object  -   instance of the Handler class
        """
        self._datasource_object = datasource_object
        self._handler_name = handler_name
        if handler_id is None:
            self._handler_id = self._get_handler_id(handler_name)
        else:
            self._handler_id = handler_id
        self.commcell_obj = self._datasource_object._commcell_object
        self._share_handler = self.commcell_obj._services['SHARE_HANDLER']

    @property
    def handler_id(self):
        """Returns the value of the handler id attribute."""
        return self._handler_id

    def _get_handler_id(self, handler_name):
        """ Get handler id for given handler name
                Args:

                    handler_name (str) -- Name of the Handler

                Returns:

                    int                -- Handler id

                Raises:

                    SDKExpception:

                        if response is empty

                        if response is not success

                """

        handlers = self.commcell_obj.Datacube.datasources.ds_handlers
        return handlers.get_properties(handler_name=handler_name)['handlerId']

    def get_handler_data(self, handler_filter=""):
        """ Executes handler for fetching data
                Args:

                     handler_filter    (str)  -- Filter which needs to applied for handler execution

                Returns:

                    dict        --  Dictionary of values fetched from handler execution

                Raises:
                    SDKExpception:

                        if response is empty

                        if response is not success

                        if error in fetching handler data
        """

        if not isinstance(handler_filter, basestring):
            raise SDKException('Datacube', '101')
        self._execute_handler = self.commcell_obj._services['EXECUTE_HANDLER'] % (
            self.handler_id, self._handler_name, handler_filter
        )
        flag, response = self.commcell_obj._cvpysdk_object.make_request(
            'GET', self._execute_handler)
        if flag:
            if response.json() and 'response' in response.json():
                return response.json()['response']
            if 'error' in response.json():
                error_message = response.json()['error']['errLogMessage']
                o_str = 'Failed to execute handler on datasource\nError: "{0}"'.format(error_message)
                raise SDKException('Datacube', '102', o_str)
            raise SDKException('Datacube', '102', "No response object in Json")
        raise SDKException('Response', '101', response.text)

    def share(self, permission_list, operation_type, user_id, user_name, user_type):
        """ Share handler with user/usergroup
                Args:
                    permission_list (list)      -- List of permission

                    operation_type  (int)       -- Operation type (2-add / 3- delete)

                    user_id         (int)       -- User id of share user

                    user_name       (str)       -- Share user name

                    user_type       (int)       -- Share user type (Ex : 13- User)

                Returns:

                    None

                Raises:

                     SDKExpception:

                        if response is empty

                        if response is not success

                        if failed to share handler with User/Usergroup
        """

        category_permission_list = []
        for permission in permission_list:
            category_permission_list.append({'permissionId': permission, '_type_': 122})
        request_json = {
            "entityAssociated": {
                "entity": [
                    {
                        "entityType": 157,
                        "_type_": 150,
                        "entityId": self.handler_id
                    }
                ]
            },
            "securityAssociations": {
                "processHiddenPermission": 1,
                "associationsOperationType": operation_type,
                "associations": [
                    {
                        "userOrGroup": [
                            {
                                "userId": user_id,
                                "_type_": user_type,
                                "userName": user_name
                            }
                        ],
                        "properties": {
                            "categoryPermission": {
                                "categoriesPermissionList": category_permission_list
                            }
                        }
                    }
                ]
            }
        }
        flag, response = self.commcell_obj._cvpysdk_object.make_request(
            'POST', self._share_handler, request_json)
        if flag:
            if 'response' in response.json():
                resp = response.json()['response']
                resp = resp[0]
                if resp.get('errorCode') != 0:
                    error_message = resp['errorString']
                    o_str = 'Failed to share handler on datasource\nError: "{0}"'.format(error_message)
                    raise SDKException('Datacube', '102', o_str)
                return response.json()['response']
            raise SDKException('Datacube', '102', "Empty Response")
        raise SDKException('Response', '101', response.text)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Lotus Notes Database Agent Instance.

LNDOCInstance is the only class defined in this file.

LNDOCInstance:
    _commonoption_restore_json          --  setter for  the Common options in restore JSON

    restore_in_place()                  --  performs an in place restore of the subclient

    restore_out_of_place()              --  performs an out of place restore of the subclient

"""

from __future__ import unicode_literals

from .lninstance import LNInstance
from ...exception import SDKException


class LNDMInstance(LNInstance):
    """Derived class from LNInstance Base class, representing an LNDOC instance,
        and to perform operations on that instance."""

    def _commonoption_restore_json(self, value):
        """setter for  the Common options in restore JSON"""

        if not isinstance(value, dict):
            raise SDKException('Subclient', '101')

        self._commonoption_restore_json = {
            "append": value.get('common_options_dict').get(
                'append', False
            ),
            "skip": value.get('common_options_dict').get(
                'skip', False
            ),
            "unconditionalOverwrite": value.get('common_options_dict').get(
                'unconditionalOverwrite', True
            ),
            "restoreOnlyStubExists": value.get('common_options_dict').get(
                'restoreOnlyStubExists', False
            ),
            "onePassRestore": value.get('common_options_dict').get(
                'onePassRestore', False
            ),
            "offlineMiningRestore": value.get('common_options_dict').get(
                'offlineMiningRestore', False
            ),
            "clusterDBBackedup": value.get('common_options_dict').get(
                'clusterDBBackedup', False
            ),
            "recoverToRecoveredItemsFolder": value.get('common_options_dict').get(
                'recoverToRecoveredItemsFolder', False
            ),
            "restoreToDisk": value.get('common_options_dict').get(
                'restoreToDisk', False
            ),
            "syncRestore": value.get('common_options_dict').get(
                'syncRestore', False
            ),
            "restoreToExchange": value.get('common_options_dict').get(
                'restoreToExchange', False
            ),
            "copyToObjectStore": value.get('common_options_dict').get(
                'copyToObjectStore', False
            )
        }

    def restore_in_place(
            self,
            paths,
            overwrite=True,
            restore_data_and_acl=True,
            copy_precedence=None,
            from_time=None,
            to_time=None,
            common_options_dict=None):
        """Restores the files/folders specified in the input paths list to the same location.

            Args:
                paths                   (list)  --  list of full paths of files/folders to restore

                overwrite               (bool)  --  unconditional overwrite files during restore
                    default: True

                restore_data_and_acl    (bool)  --  restore data and ACL files
                    default: True

                copy_precedence         (int)   --  copy precedence value of storage policy copy
                    default: None

                from_time           (str)       --  time to retore the contents after
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_time           (str)         --  time to retore the contents before
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                common_options_dict (dict)          -- dictionary for all the common options
                    options:
                        append                      :   append documents to the database

                            default: False

                        skip                        :   skip if already present

                            default: False

                        unconditionalOverwrite      :   overwrite the documents

                            default: False

                        restoreOnlyStubExists       :   restore only if it is a stub

                            default: False
            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if paths is not a list

                    if failed to initialize job

                    if response is empty

                    if response is not success
        """
        return super(LNDMInstance, self).restore_in_place(
            paths,
            overwrite,
            restore_data_and_acl,
            copy_precedence,
            from_time,
            to_time,
            **kwargs)

    def restore_out_of_place(
            self,
            client,
            destination_path,
            paths,
            overwrite=True,
            restore_data_and_acl=True,
            copy_precedence=None,
            from_time=None,
            to_time=None,
            common_options_dict=None):
        """Restores the files/folders specified in the input paths list to the input client,
            at the specified destionation location.

            Args:
                client                (str/object) --  either the name of the client or
                the instance of the Client

                destination_path      (str)        --  full path of the restore location on client

                paths                 (list)       --  list of full paths of
                files/folders to restore

                overwrite             (bool)       --  unconditional overwrite files during restore

                    default: True

                restore_data_and_acl  (bool)       --  restore data and ACL files

                    default: True

                copy_precedence         (int)      --  copy precedence value of storage policy copy

                    default: None

                from_time           (str)          --  time to retore the contents after

                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_time           (str)            --  time to retore the contents before

                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                 common_options_dict (dict)          -- dictionary for all the common options
                    options:
                        append                      :   append documents to the database

                            default: False

                        skip                        :   skip if already present

                            default: False

                        unconditionalOverwrite      :   overwrite the documents

                            default: False

                        restoreOnlyStubExists       :   restore only if it is a stub

                            default: False

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if client is not a string or Client instance

                    if destination_path is not a string

                    if paths is not a list

                    if failed to initialize job

                    if response is empty

                    if response is not success
        """
        return super(LNDMInstance, self).restore_out_of_place(
            client,
            destination_path,
            paths,
            overwrite,
            restore_data_and_acl,
            copy_precedence,
            from_time,
            to_time,
            **kwargs)




# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing operations on content analyzers, and a single content analyzer cloud in the commcell.

`ContentAnalyzers`, and `ContentAnalyzer` are 2 classes defined in this file.

ContentAnalyzers:    Class for representing all the Content analyzers in the commcell.

ContentAnalyzer:     Class for representing a single content analyzer cloud in the commcell.


ContentAnalyzers:

    __init__(commcell_object)           --  initialise object of the ContentAnalyzers class

     _response_not_success()            --  parses through the exception response, and raises SDKException

    refresh()                           --  refresh the content analyzers associated with the commcell

    get()                               --  Returns an instance of ContentAnalyzer class for the given CAcloud name

    get_properties()                    --  Returns the properties for the given content analyzer cloud name

    _get_all_contentanalyzers()         --  Returns dict consisting all content analyzers associated with commcell

    _get_cloud_from_collections()       --  gets all the content analyzer details from collection response

    has_cloud()                         --  Checks whether given CA cloud exists in commcell or not

    create()                            --  Creates the content analyzer cloud for the given client name

    delete()                            --  deletes the content analyzer cloud for the given cloud name

ContentAnalyzer:

    __init__(
        commcell_object,
        cloud_name,
        cloud_id=None)                  --  initialize an object of ContentAnalyzer Class with the given CACloud
                                                name and id associated to the commcell

    refresh()                           --  refresh the properties of the CAcloud

    _get_cloud_id()                     --  Gets content analyzer cloud id for the given CA cloud name

    _get_cloud_properties()             --  Gets all the details of associated content analyzer cloud


ContentAnalyzer Attributes
-----------------

    **cloud_id**    --  returns the cloudid of the content analyzer

    **cloud_url**   --  returns the url of the content analyzer

"""
from past.builtins import basestring
from .exception import SDKException
from .datacube.constants import ContentAnalyzerConstants


class ContentAnalyzers(object):
    """Class for representing all the ContentAnalyzers in the commcell."""

    def __init__(self, commcell_object):
        """Initializes an instance of the ContentAnalyzers class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

            Returns:
                object  -   instance of the ContentAnalyzers class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._content_analyzers = None
        self._api_get_content_analyzer_cloud = self._services['GET_CONTENT_ANALYZER_CLOUD']
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._update_response_(response.text))

    def get_properties(self, cacloud_name):
        """Returns a properties of the specified content analyzer cloud name.

            Args:
                cacloud_name (str)  --  name of the content analyzer cloud

            Returns:
                dict -  properties for the given content analyzer cloud name


        """
        return self._content_analyzers[cacloud_name]

    def _get_all_content_analyzers(self):
        """Gets the list of all content analyzers associated with this commcell.

            Returns:
                dict    -   dictionary consisting of dictionaries, where each dictionary stores the
                                details of a single content analyzer

                    {
                        "contentAnalyzerList": [
                                {
                                    "caUrl": "",
                                     "cloudName": "",
                                     "cloudId": 0
                                },
                                {
                                      "caUrl": "",
                                      "cloudName": "",
                                      "cloudId": 0
                                }
                        ]
                    }

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._api_get_content_analyzer_cloud
        )
        if flag:
            if response.json() and 'contentAnalyzerList' in response.json():
                return self._get_cloud_from_collections(response.json())
            raise SDKException('ContentAnalyzer', '103')
        self._response_not_success(response)

    @staticmethod
    def _get_cloud_from_collections(collections):
        """Extracts all the content analyzers, and their details from the list of collections given,
            and returns the dictionary of all content analyzers.

            Args:
                collections     (list)  --  list of all collections

            Returns:
                dict    -   dictionary consisting of dictionaries, where each dictionary stores the
                                details of a single content analyzer

        """
        _cacloud = {}
        for cacloud in collections['contentAnalyzerList']:
            cacloud_dict = {}
            cacloud_dict['caUrl'] = cacloud.get('caUrl', "")
            cacloud_dict['cloudName'] = cacloud.get('cloudName', "")
            cacloud_dict['cloudId'] = cacloud.get('cloudId', 0)
            cacloud_dict['clientId'] = cacloud.get('clientId', 0)
            _cacloud[cacloud['cloudName']] = cacloud_dict
        return _cacloud

    def refresh(self):
        """Refresh the content analyzers associated with the commcell."""
        self._content_analyzers = self._get_all_content_analyzers()

    def create(self, client_name, content_analyzer_name, temp_directory):
        """Creates an content analyzer cloud within the commcell

                Args:
                    client_name            (str)    --  Name of the client where content analyzer package is installed
                    content_analyzer_name  (str)    --  name for the content analyzer cloud
                    temp_directory        (str)     --  temp location for the content extractor
                    cloud_param           (list)    --  list of custom parameters to be parsed
                                                    into the json for content analyser meta info
                                                    [
                                                        {
                                                            "name": <name>,
                                                            "value": <value>
                                                        }
                                                    ]
                Returns:
                    None

                Raises:
                    SDKException:
                        Data type of the input(s) is not valid.

                        Response was not success.

                        Response was empty.
        """
        if not isinstance(client_name, basestring) or not isinstance(content_analyzer_name, basestring):
            raise SDKException('ContentAnalyzer', '101')
        client = self._commcell_object.clients.get(client_name)
        req_json = ContentAnalyzerConstants.REQUEST_JSON
        req_json['cloudNodes'] = [
            {
                "opType": ContentAnalyzerConstants.OPERATION_ADD,
                "nodeClientEntity": {
                    "hostName": client.client_hostname,
                    "clientId": int(client.client_id),
                    "clientName": client.client_name,
                    "_type_": 3
                },
                "nodeMetaInfos": [
                    {
                        "name": "PORTNO",
                        "value": "22000"
                    },
                    {
                        "name": "JVMMAXMEMORY",
                        "value": "4096"
                    },
                    {
                        "name": "INDEXLOCATION",
                        "value": temp_directory
                    }
                ]
            }
        ]
        req_json['cloudInfoEntity']['cloudName'] = content_analyzer_name
        req_json['cloudInfoEntity']['cloudDisplayName'] = content_analyzer_name
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['INDEX_SERVER_CREATION'], req_json)
        if flag:
            if response.json():
                error_code = response.json()['genericResp']['errorCode']
                error_string = response.json()['genericResp']['errorMessage']
                if error_code == 0:
                    self.refresh()
                else:
                    o_str = 'Failed to create content analyzer cloud. Error: "{0}"'.format(
                        error_string)
                    raise SDKException('ContentAnalyzer', '102', o_str)
            raise SDKException('Response', '102')
        self._response_not_success(response)

    def delete(self, cloud_name):
        """Deletes / removes an content analyzer from the commcell

                Args:
                    cloud_name      (str)   --  cloud name of content analyzer to be deleted from the commcell

                Raises:
                    SDKException:
                        Data type of the input(s) is not valid.

                        Response was not success.

                        Response was empty.
        """
        if cloud_name is None or not isinstance(cloud_name, basestring):
            raise SDKException('ContentAnalyzer', '101')

        cloud_id = self.get(cloud_name).cloud_id
        req_json = ContentAnalyzerConstants.REQUEST_JSON
        req_json["opType"] = ContentAnalyzerConstants.OPERATION_DELETE
        req_json['cloudInfoEntity']['cloudId'] = cloud_id

        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['INDEX_SERVER_DELETION'], req_json
        )
        if flag:
            if response.json() and 'genericResp' in response.json(
            ) and 'errorCode' not in response.json()['genericResp']:
                self.refresh()
                return
            if response.json() and 'genericResp' in response.json():
                raise SDKException(
                    'Response', '102', response.json()['genericResp'].get(
                        'errorMessage', ''))
            raise SDKException('Response', '102')
        self._response_not_success(response)

    def get(self, cloud_name):
        """Returns a ContentAnalyzer object for the given CA cloud name.

            Args:
                cloud_name (str)    --  name of the Content analyzer cloud

            Returns:

                obj                 -- Object of ContentAnalyzer class

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

                    if cacloud_name is not of type string


        """
        if not isinstance(cloud_name, basestring):
            raise SDKException('ContentAnalyzer', '101')

        if self.has_cloud(cloud_name):
            cloud_id = self._content_analyzers[cloud_name]['cloudId']
            return ContentAnalyzer(self._commcell_object, cloud_name, cloud_id)
        raise SDKException('ContentAnalyzer', '102', "Unable to get ContentAnalyzer class object")

    def has_cloud(self, cloud_name):
        """Checks if a content analyzer cloud exists in the commcell with the input name.

            Args:
                cloud_name (str)    --  name of the content analyzer

            Returns:
                bool - boolean output whether the CA cloud exists in the commcell or not

            Raises:
                SDKException:
                    if type of the CA cloud name argument is not string

        """
        if not isinstance(cloud_name, basestring):
            raise SDKException('ContentAnalyzer', '101')

        return self._content_analyzers and cloud_name.lower() in map(str.lower, self._content_analyzers)


class ContentAnalyzer(object):
    """Class for performing operations on a single content analyzer cloud"""

    def __init__(self, commcell_object, cloud_name, cloud_id=None):
        """Initialize an object of the ContentAnalyzer class.

            Args:
                commcell_object     (object)    --  instance of the commcell class

                cloud_name     (str)            --  name of the content analyzer cloud

                cloud_id       (str)            --  id of the content analyzer cloud
                    default: None

            Returns:
                object  -   instance of the ContentAnalyzer class
        """
        self._commcell_object = commcell_object
        self._cloud_name = cloud_name
        self._cloud_url = None
        self._cloud_id = None
        self._client_id = None
        if cloud_id is None:
            self._cloud_id = self._get_cloud_id(cloud_name)
        else:
            self._cloud_id = cloud_id
        self.refresh()

    def _get_cloud_id(self, cloud_name):
        """ Get CA cloud id for given CA cloud name
                Args:

                    cloud_name (str)   -- Name of the content analyzer cloud

                Returns:

                    int                -- Content analyzer cloud id

        """

        return self._commcell_object.content_analyzers.get(cloud_name).cloud_id

    def _get_cloud_properties(self):
        """ Get CA cloud properties for all content analyzers cloud in the commcell
                Args:

                    None

                Returns:

                    None

        """
        content_analyzers_dict = self._commcell_object.content_analyzers.get_properties(self._cloud_name)
        self._cloud_url = content_analyzers_dict['caUrl']
        self._cloud_id = content_analyzers_dict['cloudId']
        self._client_id = content_analyzers_dict['clientId']
        return content_analyzers_dict

    @property
    def client_id(self):
        """Returns the value of the Content analyzer client id attribute."""
        return self._client_id

    @property
    def cloud_id(self):
        """Returns the value of the Content analyzer cloud id attribute."""
        return self._cloud_id

    @property
    def cloud_url(self):
        """Returns the value of the Content analyzer cloud url attribute."""
        return self._cloud_url

    def refresh(self):
        """Refresh the content analyzer details associated with this commcell"""
        self._get_cloud_properties()



"""

Main File for performing Find operation at subclient Level

"""
import os
import logger
from collections import OrderedDict


class Find(object):
	"""Class for Find operation
	Methods :
	find = Performs search based on filename provided
	"""

	def __init__(self, loggingPath, commcell_object, client_name, agent_name, backupset_name):
		"""Initialize the browse  for the given path.

			Args:
				commcell_object - instance of commcell class
				client_name - Client Name
				agent_name - default is "File System" Agent
				backupset_name
			Returns:
				Instance of Find class


		"""
		self._commcell_object = commcell_object
		self._client_name = client_name
		self._agent_name = agent_name
		self._backupset_name = backupset_name
		self._LOG = logger.CVCLogger()
		self._LOG.initializelogger('cvc.log', loggingPath)
		self._LOG = self._LOG.log
		try:
			self._clientobj = self._commcell_object.clients.get(
				str(self._client_name))
			self._LOG.info("client object created successfully")
			self._agentobj = self._clientobj.agents.get(str(self._agent_name))
			self._LOG.info("agent object created successfully")
			if self._backupset_name is None:
				self._backupset_name = self._agentobj.backupsets.default_backup_set
			self._backupsetobj = self._agentobj.backupsets.get(self._backupset_name)
			self._LOG.info("backupset object created successfully")
		except Exception as backupset_obj_exception:
			self._LOG.error("Exception while creating backupset object: %s" % str(backupset_obj_exception))
			print("Exception while creating backupset object: %s" % str(backupset_obj_exception))

	def _subclient_object_creation(self, subclientname):
		"""
		Method to create subclient object
		Args :
		Find Class object
		Returns :
		Object of subclient
		"""
		try:
			flag = 1
			self._LOG.info("Client Name : %s"% (self._client_name))
			self._LOG.info("Agent Name  :%s "% (self._agent_name))
			self._LOG.info("BackupSet Name : %s"% (self._backupset_name))
			subclientobj = self._backupsetobj.subclients.get(
				str(subclientname))
			self._LOG.info("subclient object created successfully")
			flag = 0
		except Exception as subclient_exception:
			subclient_exception = str(subclient_exception)
			self._LOG.error(
				"Exception Raised in Subclient Object creation : %s"% (subclient_exception))
			flag = 1
		finally:
			return (flag, subclientobj)

	def _find_list_display(self, browse_list, browse_dict):
		"""
		Method to print the Find List of items
		"""
		try:
			self._LOG.info(browse_list)
			complete_browse_dict = {}
			for key1, value1 in browse_dict.items():
				parent_dir = os.path.abspath(os.path.join(key1, os.pardir))
				if parent_dir not in complete_browse_dict:
					complete_browse_dict[parent_dir] = {}
				complete_browse_dict[parent_dir].update({key1:value1})
			for key2, value2 in complete_browse_dict.items():
				print("\n" + key2 + ": \n")
				self._find_display(value2)
			print()

		except Exception as e:
			find_display_exception = str(e)
			self._LOG.error("Exception Raised in Find Display : %s"%
					  (find_display_exception))

	def _find_display(self, browse_dict):
		"""
		Method to print the details of files in the path
		"""
		dflag = 0
		self._LOG.info("Printing details of files")
		try:
			print ("{:<30} {:<8} {:<20} {:<8}".format('Name','Size','MTime','Type'))
			for key1, value1 in browse_dict.items():
				print ("{:<30} {:<8} {:<20} {:<8}".format(browse_dict[key1]['name'], browse_dict[key1]['size'], browse_dict[key1]['modified_time'], browse_dict[key1]['type']))
		except Exception as e:
			find_display_exception = str(e)
			self._LOG.error("Exception Raised in Find Display : %s"%
					  (find_display_exception))
			dflag = 1
		finally:
			return dflag


	def find(self, filename, subclientname=None, fromtime=0, totime=0):
		"""This method is responsible for initiating find operation based on filename
		Args :
			Find  Class Instance
			filename
			subclientname
			fromtime
			totime
		Returns :
			Prints Find Results
			Flag : Boolean that represents the completion of Find
		"""
		try:
			bflag = 1
			self._LOG.info("Find")
			self._LOG.info("filename :%s" % filename)
			self._LOG.info("Fromtime and Totime")
			self._LOG.info(fromtime)
			self._LOG.info(totime)
			if subclientname is None:
				bs = self._backupsetobj.backupset_name
				bs = str(bs)
				self._LOG.info("Proceeding with Find at backupsetlevel : %s" % bs)
				(find_list, find_dict) = self._backupsetobj.find(
					file_name=filename, from_time=fromtime, to_time=totime)
				self._find_list_display(find_list, find_dict)
				bflag = 0
			else:
				self._LOG.info("Verify if this subclient exists in given backupset")
				status = self._backupsetobj.subclients.has_subclient(
					subclientname)
				if status is True:
					subclientobj = self._backupsetobj.subclients.get(
						str(subclientname))
					self._LOG.info("subclient object created successfully")
					self._LOG.info("Path mentioned in command : %s"% (filename))
					(find_list, find_dict) = subclientobj.find(
						file_name=filename, from_time=fromtime, to_time=totime)
					self._find_list_display(find_list, find_dict)
					bflag = 0
				else:
					print("Given subclient does not exists in backupset :%s" %
						  self._backupsetobj.backupset_name)
					self._LOG.error(
						"Given subclient does not exists in backupset.So Enter proper subclient")
		except Exception as find_exception:
			self._LOG.error("Exception occurred in Find : %s"% (find_exception))
			bflag = 1
		finally:
			if bflag == 0:
				self._LOG.info("Find Completed successfully")
			else:
				self._LOG.error(" Find Encountered Error")
			return bflag



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""
        Module for operating on a Big Data Apps Subclient

        BigDataAppsSubclient:

            __init__()                  --  Just inializes all properties related to its super class

            set_data_access_nodes(data_access_nodes) -- adds the passed json object as data access
                                                        nodes for this subclient.

"""

from __future__ import unicode_literals
from ..subclients.fssubclient import FileSystemSubclient
from ..exception import SDKException


class BigDataAppsSubclient(FileSystemSubclient):
    """
        Derived class from FileSystemSubclient. Can perform fs subclient operations.
    """
    def __new__(cls, backupset_object, subclient_name, subclient_id=None):
        """
        Object creation function for BigDataAppsSubclient which returns appropiate
        sub class object based on cluster type

        Args:
            backupset_object    (obj)   --  Backupset object associated with the
            subclient

            subclient_name      (str)   --  Subclient name

            subclient_id        (int)   --  Subclient Id

        Returns:
            object              (obj)   --  Object associated with the Bigdatapps subclient

        """
        from ..subclients.splunksubclient import SplunkSubclient
        from ..subclients.index_server_subclient import IndexServerSubclient
        cluster_types = {
            16: SplunkSubclient,
            6: IndexServerSubclient
        }

        bigdata_apps_cluster_type = backupset_object._instance_object.properties. \
            get('distributedClusterInstance', {}).get('clusterType', -1)

        if bigdata_apps_cluster_type in cluster_types.keys():
            cluster_type = cluster_types[bigdata_apps_cluster_type]
            return object.__new__(cluster_type)

        return object.__new__(cls)

    def set_data_access_nodes(self, data_access_nodes):
        """
            Sets the Data Access Nodes for the distributed apps subclient.
            Args :

                data_access_nodes (list) : Sets the list of client nodes passed as
                                            data access node for this distributed apps
                                            subclient

            Raise SDK Exception :

                If unable to set data access nodes property of the subclient.

        """

        data_access_nodes_client_json = []
        for access_node in data_access_nodes:
            data_access_nodes_client_json.append({"clientName": access_node})

        data_access_nodes_json = {
            "dataAccessNodes": data_access_nodes_client_json
        }

        request_json = {
            "subClientProperties": {
                "dfsSubclientProp": {
                    "distributedDataAccessNodes": data_access_nodes_json
                }
            }
        }

        flag, response = self._cvpysdk_object.make_request(
            'POST', self._SUBCLIENT, request_json)

        output = self._process_update_response(flag, response)

        if output[0]:
            return

        o_str = 'Failed to update properties of subclient\nError: "{0}"'
        raise SDKException('Subclient', '102', o_str.format(output[2]))



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""" Main file for performing the download operation

Download
========

    __init__(commcell_object)        --  initialize commcell_object of Install class
    associated with the commcell

    repair_software                -- triggers Repair of the software on a specified
                                        client/client group

    push_servicepack_and_hotfix()    --  installs the latest service pack in the client machine

    install_software                 --  Installs the features selected on the machines selected

"""

from ..job import Job
from ..exception import SDKException
from ..deployment.deploymentconstants import UnixDownloadFeatures, WindowsDownloadFeatures


class Install(object):
    """"class for installing software packages"""

    def __init__(self, commcell_object):
        """Initialize object of the Install class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

            Returns:
                object - instance of the Install class

        """

        self.commcell_object = commcell_object
        self._services = commcell_object._services
        self._cvpysdk_object = commcell_object._cvpysdk_object

    def repair_software(self,
                        client=None,
                        client_group=None,
                        username=None,
                        password=None,
                        reboot_client=False):
        """triggers Repair of the software for a specified client machine

                Args:
                    client (str)               -- Client machine to re-install service pack on

                    client_group (str)         -- Client group to re-install service pack on
                                                            (eg : 'Media Agent')

                    username    (str)               -- username of the machine to re-install features on

                        default : None

                    password    (str)               -- base64 encoded password

                        default : None

                    reboot_client (bool)            -- boolean to specify whether to reboot the client
                    or not

                        default: False

                Returns:
                    object - instance of the Job class for this download job

                Raises:
                        SDKException:
                        if re-install job failed

                        if response is empty

                        if response is not success

        **NOTE:** repair_software can be used for client/ client_group not both; When both inputs are given only the
                  client computer will be repaired

        **NOTE:** If machine requires reboot and reboot is not selected, machine won't be updated

        **NOTE:** If machine requires login credentials and if not provided - client reinstallation might fail.

        """
        if (client is None) and (client_group is None):
            raise SDKException('Install', '100')

        if client:
            client_group = ""
            if not client in self.commcell_object.clients.all_clients:
                raise SDKException('Install', '101')

        elif client_group:
            client = ""
            if not client_group in self.commcell_object.client_groups.all_clientgroups:
                raise SDKException('Install', '102')

        request_json = {
            "taskInfo": {
                "task": {
                    "taskType": 1,
                    "initiatedFrom": 2,
                    "policyType": 0,
                    "taskFlags": {
                        "disabled": False
                    }
                },
                "subTasks": [
                    {
                        "subTaskOperation": 1,
                        "subTask": {
                            "subTaskType": 1,
                            "operationType": 4020
                        },
                        "options": {
                            "adminOpts": {
                                "clientInstallOption": {
                                    "installerOption": {
                                        "clientComposition": [
                                            {
                                                "packageDeliveryOption": 0
                                            }
                                        ]
                                    }
                                },
                                "updateOption": {
                                    "installUpdateOptions": 0,
                                    "restartExplorerPlugin": True,
                                    "rebootClient": reboot_client,
                                    "clientAndClientGroups": [
                                        {
                                            "clientGroupName": client_group,
                                            "clientName": client
                                        }
                                    ],
                                    "installUpdatesJobType": {
                                        "installType": 4,
                                        "upgradeClients": False,
                                        "undoUpdates": False,
                                        "installUpdates": False
                                    }
                                }
                            }
                        }
                    }
                ]
            }
        }

        if username:
            request_json["taskInfo"]["subTasks"][0]["options"]["adminOpts"]["clientInstallOption"]["clientAuthForJob"] \
                = {
                    "password": password,
                    "userName": username
                }

        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CREATE_TASK'], request_json
        )

        if flag:
            if response.json():
                if "jobIds" in response.json():
                    return Job(self.commcell_object, response.json()['jobIds'][0])

                else:
                    raise SDKException('Install', '107')

            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101')

    def push_servicepack_and_hotfix(
            self,
            client_computers=None,
            client_computer_groups=None,
            all_client_computers=False,
            all_client_computer_groups=False,
            reboot_client=False,
            run_db_maintenance=True,
            maintenance_release_only=False):
        """Installs the software packages on the clients

        Args:
            client_computers (list)               -- Client machines to install service pack on

            client_computer_groups (list)         -- Client groups to install service pack on

            all_client_computers (bool)           -- boolean to specify whether to install on
            all client computers or not

                default: False

            all_client _computer_groups (bool)    -- boolean to specify whether to install on all
            client computer groups or not

                default: False

            reboot_client (bool)                  -- boolean to specify whether to reboot the
            client or not

                default: False

            run_db_maintenance (bool)             -- boolean to specify whether to run db
            maintenance not

                default: True

            maintenance_release_only (bool)       -- for clients of feature releases lesser than CS, this option
            maintenance release of that client FR, if present in cache

        Returns:
            object - instance of the Job class for this download job

        Raises:
                SDKException:
                    if Download job failed

                    if response is empty

                    if response is not success

                    if another download job is already running

        **NOTE:** push_serivcepack_and_hotfixes cannot be used for revision upgrades

        """
        selected_clients = []
        selected_client_groups = []

        if not any([all_client_computers,
                    all_client_computer_groups,
                    client_computers,
                    client_computer_groups]):
            raise SDKException('Install', '101')

        commcell_client_computers = self.commcell_object.clients.all_clients
        commcell_client_computer_groups = self.commcell_object.client_groups.all_clientgroups

        if client_computers is not None:
            client_computers = [x.lower() for x in client_computers]
            if not set(client_computers).issubset(commcell_client_computers):
                raise SDKException('Install', '102')

            selected_clients = [{'clientName': client} for client in client_computers]

        if client_computer_groups is not None:
            client_computer_groups = [x.lower() for x in client_computer_groups]
            if not set(client_computer_groups).issubset(commcell_client_computer_groups):
                raise SDKException('Install', '103')

            selected_client_groups = [{'clientGroupName': client}
                                      for client in client_computer_groups]

        if all_client_computers:
            selected_clients = [{"_type_": 2}]

        if all_client_computer_groups:
            selected_client_groups = [{"_type_": 27}]

        all_clients = selected_clients + selected_client_groups

        request_json = {
            "taskInfo": {
                "task": {
                    "taskType": 1,
                    "initiatedFrom": 2,
                    "policyType": 0,
                    "alert": {
                        "alertName": ""
                    },
                    "taskFlags": {
                        "isEdgeDrive": False,
                        "disabled": False
                    }
                },
                "subTasks": [
                    {
                        "subTaskOperation": 1,
                        "subTask": {
                            "subTaskType": 1,
                            "operationType": 4020
                        },
                        "options": {
                            "adminOpts": {
                                "updateOption": {
                                    "removeIntersectingDiag": True,
                                    "restartExplorerPlugin": True,
                                    "rebootClient": reboot_client,
                                    "runDBMaintenance": run_db_maintenance,
                                    "maintenanceReleaseOnly": maintenance_release_only,
                                    "clientAndClientGroups": all_clients,
                                    "installUpdatesJobType": {
                                        "upgradeClients": False,
                                        "undoUpdates": False,
                                        "installUpdates": True
                                    }
                                }
                            },
                        }
                    }
                ]
            }
        }
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CREATE_TASK'], request_json
        )

        if flag:
            if response.json():
                if "jobIds" in response.json():
                    return Job(self.commcell_object, response.json()['jobIds'][0])

                else:
                    raise SDKException('Install', '107')

            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101')

    def install_software(
            self,
            client_computers=None,
            windows_features=None,
            unix_features=None,
            username=None,
            password=None,
            install_path=None,
            log_file_loc=None,
            client_group_name=None,
            storage_policy_name=None,
            sw_cache_client=None,
            **kwargs):
        """
        Installs the features selected on the given machines
        Args:

            client_computers    (list)      -- list of hostnames/IP address to install the
            features on

                default : None

            windows_features (list of enum) -- list of windows features to be installed

                default : None

            unix_features (list of enum)    -- list of unix features to be installed

                default : None

            username    (str)               -- username of the machine to install features on

                default : None

            password    (str)               -- base64 encoded password

                default : None

            install_path (str)              -- Install to a specified path on the client

                 default : None

            log_file_loc (str)              -- Install to a specified log path on the client

                 default : None

            client_group_name (list)        -- List of client groups for the client

                 default : None

            storage_policy_name (str)       -- Storage policy for the default subclient

                 default : None

            sw_cache_client (str)           -- Remote Cache Client Name/ Over-riding Software Cache

                default : None (Use CS Cache by default)

            **kwargs: (dict) -- Key value pairs for supporting conditional initializations
            Supported -
            install_flags (dict) - dictionary of install flag values
            Ex : install_flags = {"preferredIPFamily":2, "install32Base":True}

            db2_logs_location (dict) - dictionary of db2 logs location
            Ex: db2_logs_location = {
                                    "db2ArchivePath": "/opt/Archive/",
                                    "db2RetrievePath": "/opt/Retrieve/",
                                    "db2AuditErrorPath": "/opt/Audit/"
                            }

        Returns:
                object - instance of the Job class for this install_software job

        Raises:
            SDKException:
                if install job failed

                if response is empty

                if response is not success

        Usage:

            -   UnixDownloadFeatures and WindowsDownloadFeatures enum is used for providing
                input to the install_software method, it can be imported by

                >>> from cvpysdk.deployment.deploymentconstants import UnixDownloadFeatures
                    from cvpysdk.deployment.deploymentconstants import WindowsDownloadFeatures

            -   sample method call

                >>> commcell_obj.install_software(
                                client_computers=[win_machine1, win_machine2],
                                windows_features=[WindowsDownloadFeatures.FILE_SYSTEM.value],
                                unix_features=None,
                                username='username',
                                password='password',
                                install_path='C:\\Temp,
                                log_file_loc='/var/log',
                                client_group_name=[My_Servers],
                                storage_policy_name='My_Storage_Policy',
                                install_flags={"preferredIPFamily":2})

                    **NOTE:** Either Unix or Windows clients_computers should be chosen and
                    not both

        """
        db2_install = False
        if windows_features:
            os_type = 0
            if WindowsDownloadFeatures.DB2_AGENT.value in windows_features:
                db2_install = True
            install_options = [{'osType': 'Windows', 'ComponentId': feature_id}
                               for feature_id in windows_features]

        elif unix_features:
            os_type = 1
            if UnixDownloadFeatures.DB2_AGENT.value in unix_features:
                db2_install = True

            install_options = [{'osType': 'Unix', 'ComponentId': feature_id}
                               for feature_id in unix_features]

        else:
            raise SDKException('Install', '105')

        if client_computers:
            commcell_name = self.commcell_object.commserv_name

            client_details = []
            for client_name in client_computers:
                client_details.append(
                    {
                        "clientEntity": {
                            "clientName": client_name,
                            "commCellName": commcell_name
                        }
                    })

        else:
            raise SDKException('Install', '106')

        if client_group_name:
            client_group_name = [x.lower() for x in client_group_name]
            if not set(client_group_name).issubset(self.commcell_object.client_groups.all_clientgroups):
                raise SDKException('Install', '103')
            selected_client_groups = [{'clientGroupName': client_group}
                                      for client_group in client_group_name]

        install_flags = kwargs.get('install_flags')
        db2_logs = kwargs.get('db2_logs_location', {})

        request_json = {
            "taskInfo": {
                "associations": [
                    {
                        "commCellId": 2
                    }
                ],
                "task": {
                    "taskType": 1,
                    "initiatedFrom": 1,
                    "taskFlags": {
                        "disabled": False
                    }
                },
                "subTasks": [
                    {
                        "subTask": {
                            "subTaskType": 1,
                            "operationType": 4026
                        },
                        "options": {
                            "adminOpts": {
                                "clientInstallOption": {
                                    "reuseADCredentials": False,
                                    "installOSType": os_type,
                                    "discoveryType": 0,
                                    "installerOption": {
                                        "requestType": 0,
                                        "Operationtype": 0,
                                        "CommServeHostName":
                                            self.commcell_object.commserv_hostname,
                                        "RemoteClient": False,
                                        "installFlags": {
                                            "allowMultipleInstances": True,
                                            "restoreOnlyAgents": False,
                                            "killBrowserProcesses": True,
                                            "install32Base": install_flags.get('install32Base', False) if install_flags else False,
                                            "disableOSFirewall": False,
                                            "stopOracleServices": False,
                                            "skipClientsOfCS": False,
                                            "addToFirewallExclusion": True,
                                            "ignoreJobsRunning": False,
                                            "forceReboot": False,
                                            "overrideClientInfo": True,
                                            "preferredIPFamily": install_flags.get('preferredIPFamily', 1) if install_flags else 1,
                                            "firewallInstall": {
                                                "enableFirewallConfig": False,
                                                "firewallConnectionType": 0,
                                                "portNumber": 0
                                            }
                                        },
                                        "User": {
                                            "userName": "admin",
                                            "userId": 1
                                        },
                                        "clientComposition": [
                                            {
                                                "overrideSoftwareCache": True if sw_cache_client else False,
                                                "softwareCacheOrSrmProxyClient": {
                                                    "clientName": sw_cache_client if sw_cache_client else ""
                                                },
                                                "packageDeliveryOption": 0,
                                                "components": {
                                                    "commonInfo": {
                                                        "globalFilters": 2,
                                                        "storagePolicyToUse": {
                                                            "storagePolicyName": storage_policy_name if storage_policy_name else ""
                                                        }
                                                    },
                                                    "fileSystem": {
                                                        "configureForLaptopBackups": False
                                                    },
                                                    "componentInfo": install_options,
                                                },
                                                "clientInfo": {
                                                    "clientGroups": selected_client_groups if client_group_name else [],
                                                    "client": {
                                                        "evmgrcPort": 0,
                                                        "cvdPort": 0,
                                                        "installDirectory": install_path if install_path else ""
                                                    },
                                                    "clientProps": {
                                                        "logFilesLocation": log_file_loc if log_file_loc else ""
                                                    }
                                                }
                                            }
                                        ]
                                    },
                                    "clientDetails": client_details,
                                    "clientAuthForJob": {
                                        "password": password,
                                        "userName": username
                                    }
                                },
                                "updateOption": {
                                    "rebootClient": True
                                }
                            }
                        }
                    }
                ]
            }
        }

        if db2_install and db2_logs:
            request_json["taskInfo"]["subTasks"][0]["options"]["adminOpts"]["clientInstallOption"]["installerOption"][
                "clientComposition"][0]["components"]["db2"] = db2_logs

        flag, response = self._cvpysdk_object.make_request(
            'POST', self._services['CREATE_TASK'], request_json
        )

        if flag:
            if response.json():
                if "jobIds" in response.json():
                    return Job(self.commcell_object, response.json()['jobIds'][0])

                else:
                    raise SDKException('Install', '107')

            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101')



"""static

Main File for performing browse operation at subclient Level adn bacupset level.

"""

import datetime
import os
import platform
import subprocess
import traceback
import json
from requests import Response
import xmltodict

from entity import Entity


class Browse(Entity):
    """ Class for browse operation """
    def __init__(
            self,
            package_instance,
            commcell_object,
            client_name,
            agent_name,
            backupset_name,
            logger_object):
        """ Initialize the browse class instance

            Args:
                package_instance   (str)   -- Instance where commvault is installed

                commcell_object (obj)   -- instance of commcell class

                client_name     (str)   -- Client to perform browse on

                agent_name      (str)   -- Agent to perform browse on

                backupset_name  (str)   -- Backupset to be used for browse

                logger_object   (str)   -- Instance of Logger class

        """
        self._LOG = logger_object
        self.live_browse_path = ''
        self.agent_name = agent_name
        super(Browse, self).__init__(
            logger_object=logger_object,
            commcell_object=commcell_object,
            client_name=client_name,
            agent_name=agent_name,
            backupset_name=backupset_name,
            package_instance=package_instance
        )

    def _browse_display_versions(self, browse_dict):
        """
        Method to print the details of different verions of a file
        """
        dflag = 0
        self._LOG.info("Printing details of versions")
        try:
            if browse_dict == ([], {}):  #This is the return for empty browse with -allVersions
                print("Browse versions returned empty, please check whether the user has the required permissions")
                return dflag
            print("{:<30}\t{:<8}\t{:<8}\t{:<20}\t{:<8}".format('Name', 'Version', 'Size', 'MTime', 'Type'))
            print('_ ' * 50)
            print('\n')
            for item in browse_dict[next(iter(browse_dict))]:
                print("{:<30}\t{:<8}\t{:<8}\t{:<20}\t{:<8}".format(
                    item['name'] or ' ', 
                    item['version'] or ' ', 
                    item['size'] or ' ', 
                    datetime.datetime.strptime(item['modified_time'], "%d/%m/%Y %H:%M:%S").strftime("%Y-%m-%d %H:%M:%S") if item['modified_time'] else ' ',
                    item['type'] or ' '))
        except Exception as exp:
            browse_display_exception = str(exp)
            self._LOG.error("Exception Raised in Browse Versions Display: %s", browse_display_exception)
            dflag = 1
        finally:
            return dflag

    def _browse_display(self, browse_dict):
        """
        Method to print the details of files in the path
        """
        dflag = 0
        self._LOG.info("Printing details of files")
        try:
            if not browse_dict:
                print("Browse returned empty, please check whether the user has the required permissions")
                return dflag
            print("{:<30}\t{:<8}\t{:<20}\t{:<8}".format('Name', 'Size', 'MTime', 'Type'))
            print('_ ' * 40)
            print('\n')
            for key in browse_dict:
                print("{:<30}\t{:<8}\t{:<20}\t{:<8}".format(
                    browse_dict[key]['name'] or ' ',
                    browse_dict[key]['size'] or ' ',
                    datetime.datetime.strptime(browse_dict[key]['modified_time'], "%d/%m/%Y %H:%M:%S").strftime("%Y-%m-%d %H:%M:%S") if browse_dict[key]['modified_time'] else ' ',
                    browse_dict[key]['type'] or ' '))
        except Exception as exp:
            browse_display_exception = str(exp)
            self._LOG.error("Exception Raised in Browse Display : %s", browse_display_exception)
            dflag = 1
        finally:
            return dflag

    def _find_display(self, browse_dict, vsa_guid=""):
        """
        Method to print the details of files in the path
        """
        dflag = 0
        self._LOG.info("Printing details of files")
        try:
            if not browse_dict:
                print("Find returned empty, please check whether the user has the required permission")
                return dflag
            print("{:<30}\t{:<8}\t{:<20}\t{:<8}".format('File path', 'Size', 'MTime', 'Type'))
            print('_ ' * 40)
            print('\n')
            temp = ''
            for key in browse_dict:
                if vsa_guid:
                    temp_key = key.strip('\\').strip(vsa_guid).replace('\\', '/')
                    temp_key = os.path.dirname(temp_key)
                else:
                    temp_key = os.path.dirname(key)
                if temp != temp_key:
                    print(temp_key)
                    temp = temp_key
                print("\t{:<30}\t{:<8}\t{:<20}\t{:<8}".format(
                    browse_dict[key]['name'] or ' ',
                    browse_dict[key]['size'] or ' ',
                    datetime.datetime.strptime(browse_dict[key]['modified_time'], "%d/%m/%Y %H:%M:%S").strftime("%Y-%m-%d %H:%M:%S") if browse_dict[key]['modified_time'] else ' ',
                    browse_dict[key]['type'] or ' '))
        except Exception as exp:
            browse_display_exception = str(exp)
            self._LOG.error("Exception Raised in Find Display : %s", browse_display_exception)
            dflag = 1
        finally:
            return dflag

    def qlist_browse(
            self,
            operation='Browse',
            subclient_name=None,
            path=None,
            from_time=None,
            to_time=None,
            filters=None,
            vsa_guid="",
    ):
        """
        Performs browse operation using qlist

        Args:

            operation       (str)   --  Operation to be performed
                                            default: 'Browse'

                                            Accepted Values:
                                            1. Browse
                                            2. Find
                                            3. All versions

            subclient_name	(str)	--	Name of the subclient to perform browse on

            path            (str)   --  Path to be browsed

            from_time		(str)	--	From time for the browse operation

            to_time			(str)	--	To time to be set for the browse operation

            filters			(list)	--	List of filter options

            vsa_guid        (str)   --  VSA GUID used to browse VM

        Returns:
            (bool)	--	Returns True if operation is successful

        """
        try:
            flag = 0
            input_file = ''
            output_file = ''
            temp_folder = ''

            # To support relative path passed by user
            path = Entity._get_absolute_path(path, self.live_browse_path)

            if path not in ['/', '\\']:
                path = path.rstrip('/').rstrip('\\')

            options = {
                'path': path,
                'from_time': from_time,
                'to_time': to_time,
                'filters': filters or [],
            }

            os_type = platform.system()
            backupset_name = Entity.backupset.backupset_name

            if subclient_name:
                self._LOG.info(
                    "Verifying whether subclient: %s exists in backupset: %s",
                    subclient_name,
                    backupset_name
                )
                if Entity.backupset.subclients.has_subclient(subclient_name):
                    self._LOG.info("Subclient exists in backupset")
                    options['_subclient_id'] = Entity.backupset.subclients[subclient_name.lower()]['id']
                else:
                    raise Exception(
                        "Given subclient is not present in the backupset, please enter a valid subclient name")
            else:
                self._LOG.info("Proceeding with browse at backupset level : %s", backupset_name)

            if operation.lower() == 'all versions':
                self._LOG.info("Browsing all versions")
                options['operation'] = 'all_versions'
            elif operation.lower() == 'find':
                self._LOG.info("Proceeding with find operation")
                options['operation'] = 'find'
                options['path'] = path + ("\\**\\*" if 'windows' in Entity.client.os_info.lower() else "/**/*")
            else:
                self._LOG.info("Browsing latest version")
                options['operation'] = 'browse'

            if not options['path']:
                options['path'] = '\\'

            self._LOG.info('Browsing path: %s', options['path'])

            if self.agent_name == 'virtual server':
                options['vs_file_browse'] = 'true'
                options['retry_count'] = 0
                if operation.lower() == 'find':
                    options['path'] = '\\{0}\\**\\*'.format(vsa_guid.rstrip("\n"))
                else:
                    options['path'] = '\\{0}{1}'.format(vsa_guid.rstrip("\n"), options['path'].replace('/', '\\'))
                self._LOG.info("VSA Browse path is %s", options['path'])

            options = Entity.backupset._prepare_browse_options(options)
            request_json = Entity.backupset._prepare_browse_json(options)

            if os_type == 'OS400':
                response_flag, response = Entity.commcell._cvpysdk_object.make_request(
                    'POST',
                    Entity.commcell._services['BROWSE'],
                    request_json)
            else:
                # Converting dict to xml for qlist
                request_json = {'databrowse_BrowseRequest': request_json}
                request_xml = xmltodict.unparse(request_json)

                status, installation_path, error = Entity.simpana_default_obj.simpana_installation_path(
                    Entity.package_instance
                )

                if not status:
                    raise Exception(
                        "No local installation present with the specified instance name."
                        " Cannot perform browse\nError: {}".format(error))

                _time = '{date:%m-%d-%Y_%H_%M_%S}'.format(date=datetime.datetime.now())
                _time = '{}_{}'.format(_time, os.getpid())
                if os_type == 'Linux':
                    command = '{}/Base/qlist'.format(installation_path)
                    try:
                        temp_folder = r'/tmp/cvc'
                        if not os.path.exists(temp_folder):
                            os.makedirs(temp_folder)
                            os.chmod(temp_folder, 0o777)
                    except OSError:
                        temp_folder = os.getcwd()

                    input_file = '{}/cvc_browse_request_{}.xml'.format(temp_folder, _time)
                    output_file = '{}/cvc_browse_response_{}.xml'.format(temp_folder, _time)

                elif os_type == 'Windows':
                    command = r'{}\Base\qlist'.format(installation_path)

                    try:
                        temp_folder = r'{}\Temp\cvc'.format(installation_path)
                        if not os.path.exists(temp_folder):
                            os.makedirs(temp_folder)
                    except OSError:
                        temp_folder = os.getcwd()

                    input_file = r'{}\cvc_browse_request_{}.xml'.format(temp_folder, _time)
                    output_file = r'{}\cvc_browse_response_{}.xml'.format(temp_folder, _time)
                else:
                    raise Exception('OS: {} not supported'.format(os_type))

                self._LOG.debug('Input file: %s', input_file)
                self._LOG.debug('Output file: %s', output_file)

                with open(input_file, 'w') as _file:
                    _file.write(request_xml)

                command = '{} backupfiles -af "{}" -tk "{}" -dpath "{}"'.format(
                    command,
                    input_file,
                    Entity.commcell.auth_token.split(' ')[1],
                    output_file
                )

                output = subprocess.run(command, shell=True)
                if not output.returncode == 0:
                    raise Exception("Browse failed with error: {}".format(output.stderr))

                with open(output_file) as _file:
                    content = _file.read()

                # To convert XML to dict
                content = xmltodict.parse(content, attr_prefix='')['OUTPUT']['databrowse_BrowseResponse']
                content = json.loads(json.dumps(content))

                # To convert dict as request response
                response = Response()
                response._content = str({'browseResponses': content}).replace("'", '"').encode()
                response_flag = True

            if operation.lower() == 'all versions':
                browse_dict = Entity.backupset._process_browse_response(response_flag, response, options)
                if browse_dict is not None:
                    flag = self._browse_display_versions(browse_dict)
            else:
                _, browse_dict = Entity.backupset._process_browse_response(response_flag, response, options)
                if browse_dict is not None:
                    if operation.lower() == 'find':
                        flag = self._find_display(browse_dict, vsa_guid)
                    else:
                        flag = self._browse_display(browse_dict)

        except OSError:
            print('User does not have enough permission to create temp files')
            self._LOG.error(
                'User does not have enough permission to create temp files \n Error: %s',
                traceback.format_exc()
            )

        except Exception:
            self._LOG.error("Exception occurred in browse : %s", traceback.format_exc())
            flag = 1

        finally:
            if input_file and os.path.exists(input_file):
                os.remove(input_file)
            if output_file and os.path.exists(output_file):
                os.remove(output_file)

            if flag == 0:
                self._LOG.info("Browse job completed successfully")
            else:
                self._LOG.error("Browse job encountered error")
            return flag
